#!/home/moerman/cadbin/perl -w
# Maze generator
#
# This script is a maze (Dutch: doolhof) generator, which can
# generate simple mazes. There is also a more extensive version
# of this program, for multiple size mazes, graphical output
# instead of text (.png files), easy or hard settings, etc.
#
# How does it work: if you have a room with external walls only and two
# doors, there is one way from one door to the next. Note that this does
# not change if you add one wall starting from an existing wall, but not
# connecting to anything else. This way, you can keep on adding walls,
# with only one side connected, and still keep one (and only one) route
# from the input to the output. The route just will get more complex, and
# more and more false paths are added.
# 
# This is also how this program created mazes. We start with an empty
# space (by init_maze), add the external walls (init_borders), and just
# start on adding walls. This is first done at random: find a free spot
# and from there find a wall (find_free_pole), and start creating walls at
# that point (draw_segments). As we started at a free spot, there is
# always a direction to add walls.
# 
# In the last stage, we systematically scan for free points and draw walls
# (check_all_poles), just in case we have missed any areas with the random
# filling.
# 
# This is all done on the representation using the visual metaphore of
# having a grid with poles, and tying ropes (segments) between the poles
# as walls. Remember, when we place N poles, we have only N-1 ropes needed
# between the poles.
#
# Kees Moerman, April 20, 2001 (c)
#

use strict;                             # don't forget to prefix vars with my


# ----------------------------- data structures -------------------

# representation of the maze is done in two arrays; one for the
# horizontal lines and one for the vertical ones. Look at the below
# figure for the orientation and numbering of the elements. See the
# maze as a number of poles in a grid, with lines drawn between them.
#
#      0           1           2
#    
# 0    X--h[0,0]---X--h[1,0]---X    h --> actually array horlines
#      |           |           |
#   v[0,0]      v[1,0]      v[2,0]
#      |           |           |
# 1    X--h[0,1]---X--h[1,1]---X
#      |           |           |
#   v[0,1]      v[1,1]      v[2,1]
#      |           |           |
# 2    X--h[0,2]---X--h[1,2]---X


my $xmax = 30;                          # number of lines (poles: +1)
my $ymax = 10;

my @horlines;                           # note: X poles need X-1 lines 
my @verlines;                           # see above diagram for interpretation

# hide internal representation with access functions (might change later)

sub set_hor { my ($x, $y, $z) = (@_); $horlines[$x][$y] = $z; } # $z = 0/1
sub set_ver { my ($x, $y, $z) = (@_); $verlines[$x][$y] = $z; }
sub get_hor { my ($x, $y) = (@_);     return $horlines[$x][$y]; }
sub get_ver { my ($x, $y) = (@_);     return $verlines[$x][$y]; }

my ($gx, $gy);                          # global location used by some functions

# ------------------------------ initialisation ----------------------

sub init_maze                           # initialise edge arrays with zeros
{   my ($x, $y);                        # ( ) --> 
    for $x (0..$xmax)
    {   for $y (0..$ymax)
        {   set_hor($x, $y, 0);         # just fill everyting with zeros
            set_ver($x, $y, 0);
        }
    }
}

sub init_borders                        # initialise boundaries at 1
{   my ($x, $y);                        # ( ) --> 
    for $x (0..$xmax-1)                 # X poles -> X-1 lines
    {   
        set_hor($x, 0, 1);
        set_hor($x, $ymax, 1);          # set the four walls at 1
    }
    for $y (0..$ymax-1)
    {
        set_ver(0, $y, 1);
        set_ver($xmax, $y, 1);
    }
}

# --------------------------- basic subroutines ----------------------

sub rnd {   int (rand( $_[0] )); }      # random 0..N-1: ( N ) --> ( int )

sub is_pole_free                        # check if pole is standing free
{   my ($x, $y) = (@_);                 # ( x, y ) --> ( flag )
                                        # check boundaries first
    return 0 if (($x<0) || ($y<0) || ($x>=$xmax) || ($y>=$ymax));
    return 0 if get_ver($x, $y-1);      # not free on top?
    return 0 if get_hor($x-1, $y);      # not free to the left?
    return 0 if get_ver($x, $y);        # not free at bottom?
    return 0 if get_hor($x, $y);        # not free to the right?
    return 1;                           # yes, all sides free, return success
}

sub draw_next_segment                   # add one more segment (if possible)        
{   my $dir = rnd(4);                   # on current line, random direction
                                        # operate on globals $gx, $gy, $color
    for (0..1)                          # might need to go round once (starts halfway)
    {
        if (($dir & 3) == 0)            # left?
        {   if(is_pole_free($gx-1, $gy))  { set_hor($gx-1, $gy, 1); $gx--; return 1}
            $dir++;                     # failed, try next direction
        }
            
        if (($dir & 3) == 1)            # down
        {   if(is_pole_free($gx, $gy+1))  { set_ver($gx, $gy, 1);   $gy++; return 1 }
            $dir++;
        }
            
        if (($dir & 3) == 2)            # right
        {   if(is_pole_free($gx+1, $gy))  { set_hor($gx, $gy, 1);   $gx++; return 1 }
            $dir++;
        }
            
        if (($dir & 3) == 3)            # up
        {   if(is_pole_free($gx, $gy-1))  { set_ver($gx, $gy-1, 1); $gy--; return 1 }
            $dir++;
        }
    }
    return 0;                           # give up (return failure): no free space.
}

sub draw_segments                       # generate one curly line  
{   ($gx, $gy) = (@_);                  # ( x, y ) -->
                                        # set and work on globals $gx, $gy
    while( rnd($xmax)!=0)               # limit max length
    {
        last if !draw_next_segment();   # as long as we can add segments
    }
}

sub move_near_free                      # find first occupied pole to the left/right/...
{                                       # input in globals: free standing pole
    my $rand = rnd(4);                  # look in random direction

    if( $rand == 0)                     # find first free pole to left
    {   $gx-- while( is_pole_free($gx-1, $gy));
        $gx--;                          # so this one connected to walls
    }
    elsif ($rand == 1)                  # find first free pole to top
    {   $gy++ while( is_pole_free($gx, $gy+1));
        $gy++;                          # so this one connected to walls
    }
    elsif( $rand == 2)                  # first free pole to bottom
    {   $gy-- while( is_pole_free($gx, $gy-1));
        $gy--;                          # so this one connected to walls
    }
    else                                # first free pole to right
    {   $gx++ while( is_pole_free($gx+1, $gy));
        $gx++;                          # so this one connected to walls
    }
}

sub find_free_pole                      # find random free-standing pole close to wall
{   for (0..10)                         # return in globals gx, gy
    {
        $gx = rnd($xmax-1);             # (try max 10 times to find is_pole_free spot)
        $gy = rnd($ymax-1);
        if( is_pole_free($gx, $gy))     # random free pole
        {
            move_near_free();           # so find close wall (any direction is OK)
            return 1;
        }
    }
    return 0;                           # not found
}

sub check_all_poles                     # check/fill if still empty pole
{   my ($x, $y);                        # carefull: don' use $gx as global and as loop
    for $x (1..$xmax-1)                 # using exhaustive check on globals gx, gy
    {
        for $y (1..$ymax-1)
        {   $gx = $x; $gy = $y;
            if(is_pole_free($gx, $gy))  # OK, found freestanding pole
            {
                move_near_free();       # get close wall and try to connect
                draw_segments($gx, $gy);    # may go in wrong direction, though!)
            }
        }
    }
}


# ------- and the actual maze generation and printing ---------------------

sub doolhof                             # generate complete doolhof
{                                       # ( ) -->
    init_maze();                        # 
    init_borders();                     # so external borders set up

    while ( find_free_pole())           # start random filling
    {
        draw_segments( $gx, $gy);       # another chain of walls
    }
    
    check_all_poles;                    # systematic search

    set_ver(0, $ymax/2, 0);             # clear external wall doors
    set_ver($xmax, $ymax/2, 0);         # on left and right
}

sub visualise                           # make ASCII representation
{   my ($x, $y, $temp);                 # ( ) --> ( string )

    $temp = "";                         # poles are represented by '+'
    for $y (0..$ymax-1)                 # vertical walls by '-'
    {                                   # horizontals by '|'
        for $x (0..$xmax-1)             # and the suqares between poles by ' '
        {   
            $temp .= "+" . (get_hor($x, $y)?"-":" ");
        }
        $temp .= "+\n";
        for $x (0..$xmax)
        {   
            $temp .= (get_ver($x, $y)?"|":" ") . " ";
        }
        $temp .= "\n";
    }
    for $x (0..$xmax-1)
    {   
        $temp .= "+" . (get_hor($x, $ymax)?"-":" ");
    }
    return "$temp+\n";
}

# ------------------------------ 'main' program ---------------------

doolhof();                              # generate maze
print visualise();                      # and print it on standard output


#end of perl script

